-- GLOBALS: DotHaste_Saved
local dots = {
	["PRIEST"] = {
		[3] = {
			[589] = true, -- shadow word: pain
			[34914] = true, -- vampiric touch
			[2944] = true, -- devouring plague
		},
	},
	["DRUID"] = {
		[1] = {
			[5570] = true, -- insect swarm
			[8921] = true, -- moonfire
			--[93402] = true, -- sunfire
		},
	},
	["MAGE"] = {
		[2] = {
			[44457] = true, -- LB
			[11366] = true, -- pyroblast
			[44614] = true, -- FFB
			[11129] = true, -- combust
		},
	},
	--[[["WARLOCK"] = {
		[980] = true, -- bane of agony
		[30108] = true, -- UA
		[172] = true, -- corr
		[348] = true, -- immo
	},--]]
}
local mappedSpellIDs = {
	[93402] = 8921, -- sunfire to moonfire
	[92315] = 11366, -- pyroblast! (proc) to pyroblast
}
local dotBreakpoints = {
	-- defaults taken from XLS spreadsheet, unconfirmed
	["PRIEST"] = {
		[589] = {
			8.322806,
			24.973969,
			41.676512,
			58.353136,
			74.978138,
			91.632079,
			108.405711,
			124.971895,
		},
		[34914] = {
			9.990838,
			30.010840,
			49.962517,
			70.019845,
			90.053861,
			110.010515,
			129.973187,
			150.104231,
		},
		[2944] = {
			6.25111,
			18.741346,
			31.262312,
			43.781459,
			56.290709,
			68.776381,
			81.214146,
			93.735886,
			106.256460,
			118.738623,
		},
	},
	["DRUID"] = {
		-- 3/3 genesis is assumed here
		[5570] = {
			5.568757,
			16.652092,
			27.754719,
			38.937141,
			49.941264,
			61.095462,
			72.191147,
			83.402126,
			94.457965,
		},
		[8921] = {
			5.568757,
			16.652092,
			27.754719,
			38.937141,
			49.941264,
			61.095462,
			72.191147,
			83.402126,
			94.457965,
		},
		--[[[93402] = {
			5.568757,
			16.652092,
			27.754719,
			38.937141,
			49.941264,
			61.095462,
			72.191147,
			83.402126,
			94.457965,
		},--]]
	},
	["MAGE"] = {
		[44457] = {
			12.507036,
			37.520061,
			62.469546,
			87.441436,
			112.539866,
			137.435713,
		},
		[11366] = {
			12.507036,
			37.520061,
			62.469546,
			87.441436,
			112.539866,
			137.435713,
		},
		[44614] = {
			12.507036,
			37.520061,
			62.469546,
			87.441436,
			112.539866,
			137.435713,
		},
		[11129] = {
			4.986888,
			15.008639,
			25.078190,
			35.043908,
			45.032653,
			54.918692,
			64.880489,
			74.978158,
			85.013910,
			95.121989,
		},
	},
}
local _,class = UnitClass("player")
local dotMasterTable = dots[class]
if not dotMasterTable then
	print("<dothaste> Your class is not yet supported. Load aborted.")
	return
end
local pairs, next, ipairs, select, tonumber, _G, tinsert, tremove, tContains, table, math, print =
	  pairs, next, ipairs, select, tonumber, _G, tinsert, tremove, tContains, table, math, print
local GetSpellBonusDamage, MAX_SPELL_SCHOOLS, GetSpellInfo, CreateFrame, UnitExists, UnitCanAttack, UnitGUID, UnitIsUnit, UnitSpellHaste, UnitClass, GetPrimaryTalentTree = 
	  GetSpellBonusDamage, MAX_SPELL_SCHOOLS, GetSpellInfo, CreateFrame, UnitExists, UnitCanAttack, UnitGUID, UnitIsUnit, UnitSpellHaste, UnitClass, GetPrimaryTalentTree
local function spelldamage()
	local minDmg = GetSpellBonusDamage(2)
	for school=3,MAX_SPELL_SCHOOLS do
		local dmg = GetSpellBonusDamage(school)
		if dmg < minDmg then
			minDmg = dmg
		end
	end
	return minDmg
end
local dotDataSp = {}
local dotDataHaste = {}
local dotDataBreakpoint = {}
local dotTable
local display = CreateFrame("Frame",nil,UIParent)
display.hideFlags = {}
function display:UpdateVisible()
	for _,flag in pairs(self.hideFlags) do
		if flag then
			self:Hide()
			return false
		end
	end
	self:Show()
	return true
end
display:SetMovable(true)
function display:UpdateAnchor()
	self:SetPoint(self.saved.p,self.saved.n and _G[self.saved.n],self.saved.r,self.saved.x,self.saved.y)
end
display.mover = CreateFrame("Frame",nil,UIParent)
display.mover:SetFrameLevel(display:GetFrameLevel()+1)
display.mover:SetAllPoints(display)
display.mover:SetScript("OnMouseDown",function() display:StartMoving() end)
display.mover:SetScript("OnMouseUp",function() display:StopMovingOrSizing() local s,n = display.saved s.p,n,s.r,s.x,s.y = display:GetPoint() s.n = n and n:GetName() end)
display.mover:EnableMouse(true)
display.mover.tex = display.mover:CreateTexture()
display.mover.tex:SetAllPoints()
display.mover.tex:SetAlpha(0.5)
display.mover.tex:SetTexture(0.5,0.5,0.8)
display.mover.lab = display.mover:CreateFontString()
display.mover.lab:SetPoint("CENTER")
display.mover.lab:SetFont("Fonts\\FRIZQT__.TTF", 12, "OUTLINE")
display.mover.lab:SetText("DotHaste:\n- Drag to move\n- /dothaste lock")
function display:UpdateLock()
	self.mover[self.saved.lock and "Hide" or "Show"](self.mover)
end
local slashKeyCache = {}
local slashTable = {
	lock={
		type="handler",
		handler=function()
			display.saved.lock = not display.saved.lock
			display:UpdateLock()
		end,
	},
	scale={
		type="handler",
		handler=function(_,n)
			n = tonumber(n)
			if n then
				display:SetScale(n*0.5)
				display.saved.s = n
			end
		end,
	},
	nudge = {
		type="group",
		args={
			up = {
				type="handler",
				handler=function(_,n)
					n = tonumber(n)
					if n then
						display.saved.y = display.saved.y+n
						display:UpdateAnchor()
					end
				end,
			},
			down = {
				type = "handler",
				handler=function(_,n)
					n = tonumber(n)
					if n then
						display.saved.y = display.saved.y-n
						display:UpdateAnchor()
					end
				end,
			},
			left = {
				type = "handler",
				handler=function(_,n)
					n = tonumber(n)
					if n then
						display.saved.x = display.saved.x-n
						display:UpdateAnchor()
					end
				end,
			},
			right = {
				type = "handler",
				handler = function(_,n)
					n = tonumber(n)
					if n then
						display.saved.x = display.saved.x+n
						display:UpdateAnchor()
					end
				end,
			},
		},
	},
	breakpoints={
		type = "group",
		handler = function(command, arg)
			local spell, sub = command:match("(%S+) (%S+)$")
			spell = slashKeyCache[spell]
			arg = tonumber(arg)
			arg = arg and tonumber(("%.2f"):format(arg))
			if sub == "add" then
				if not arg then
					print("Proper usage: /bh",command," <percentage>")
					return
				end
				if not display.saved.breakpoints[spell] then
					display.saved.breakpoints[spell] = {}
				end
				if tContains(display.saved.breakpoints[spell],arg) then
					print("Already defined breakpoint:",arg,"%")
				else
					tinsert(display.saved.breakpoints[spell],arg)
					table.sort(display.saved.breakpoints[spell])
					print("Added:",arg,"%")
				end
			elseif sub == "list" then
				print("Breakpoints for",spell)
				for num,pct in ipairs(display.saved.breakpoints[spell] or {}) do
					print(("%d: %.2f%%"):format(num,pct))
				end
			elseif sub == "delete" then
				if not arg then
					print("Proper usage: /bh",command," <percentage>")
					return
				end
				for key,num in ipairs(display.saved.breakpoints[spell] or {}) do
					if num == arg then
						tremove(display.saved.breakpoints[spell],key)
						print("Removed:",arg,"%")
						return
					end
				end
				print("Breakpoint not found:",arg,"%")
			end
		end,
		args={},
	},
}
local hadInit
local function init()
	hadInit = true
	display.bars = {}
	function display:Update()
		if not (UnitExists("target") and UnitCanAttack("player","target")) then
			self.hideFlags.target = true
			self:UpdateVisible()
			return
		end
		self.hideFlags.target = nil
		if not self:UpdateVisible() then
			return
		end
		local targetGUID = UnitGUID("target")
		local currentSP, currentHastePct = spelldamage(), UnitSpellHaste("player")
		for _,bar in ipairs(self.bars) do
			local id = bar.spellID
			local sp, haste, nextBreak = dotDataSp[id][targetGUID], dotDataHaste[id][targetGUID], dotDataBreakpoint[id][targetGUID]
			if sp and (sp >= currentSP) then
				bar.spLabel:SetTextColor(0.4,1,0.4)
				bar.spLabel:SetText(sp)
			elseif sp then
				bar.spLabel:SetTextColor(1,1,0.4)
				bar.spLabel:SetText(currentSP)
			else
				bar.spLabel:SetText()
			end
			if nextBreak and (currentHastePct > nextBreak) then
				bar.hasteLabel:SetTextColor(1,0.4,0.4)
				bar.spLabel:SetTextColor(1,0.4,0.4)
				bar.hasteLabel:SetText(("%.2f%%"):format(currentHastePct))
			elseif haste and (haste >= currentHastePct) then
				bar.hasteLabel:SetTextColor(0.4,1,0.4)
				bar.hasteLabel:SetText(("%.2f%%"):format(haste))
			elseif haste then
				bar.hasteLabel:SetTextColor(1,1,0.4)
				bar.hasteLabel:SetText(("%.2f%%"):format(currentHastePct))
			else
				bar.hasteLabel:SetText()
			end
		end
	end
	display:Update()
	if dotTable then
		display:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	end
	display:RegisterEvent("UNIT_SPELL_HASTE")
	display:RegisterEvent("COMBAT_RATING_UPDATE")
	display:RegisterEvent("PLAYER_DAMAGE_DONE_MODS")
	display:RegisterEvent("PLAYER_TARGET_CHANGED")
	display:RegisterEvent("UNIT_DYNAMIC_FLAGS")
	display:RegisterEvent("UNIT_FLAGS")
	display:RegisterEvent("UNIT_FACTION")
	display:RegisterEvent("UNIT_TARGETABLE_CHANGED")
end
local function talentupdate()
	local currentSpec = GetPrimaryTalentTree()
	if not currentSpec then
		print("<dothaste> Talents not found! Delaying load!")
		return
	end
	dotTable = dotMasterTable[currentSpec]
	if not dotTable then
		if not display.hideFlags.spec then
			print("<dothaste> Your current spec is not supported by DotHaste. Hiding display.")
			display.hideFlags.spec = true
			display:UpdateVisible()
		end
		display:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
		return
	end
	display:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	display.hideFlags.spec = nil
	display:UpdateVisible()
	local keyCount = 0
	table.wipe(slashTable.breakpoints.args)
	for key in pairs(dotTable) do
		local name,_,icon = GetSpellInfo(key)
		if not icon then
			display:RegisterEvent("SPELLS_CHANGED")
			return
		else
			dotDataSp[key] = dotDataSp[key] or {}
			dotDataHaste[key] = dotDataHaste[key] or {}
			dotDataBreakpoint[key] = dotDataBreakpoint[key] or {}
			dotTable[key] = icon
			name = name:gsub("[^a-zA-Z]",""):lower()
			slashTable.breakpoints.args[name] = {
				type="group",
				args={
					add={
						type="handler",
					},
					list={
						type="handler",
					},
					delete={
						type="handler",
					},
				},
			}
			slashKeyCache[name] = key
			keyCount = keyCount+1
		end
	end
	display:SetSize(250,50*keyCount)
	if not hadInit then init() end
	local lastKey
	for num=1,keyCount do
		local displayBar = display.bars[num]
		if not displayBar then
			displayBar = CreateFrame("Frame",nil,display)
			displayBar:SetSize(250,50)
			displayBar:SetPoint("TOPLEFT",0,-50*(num-1))
			displayBar.icon = displayBar:CreateTexture()
			displayBar.icon:SetPoint("TOPLEFT")
			displayBar.icon:SetSize(50,50)
			displayBar.spLabel = displayBar:CreateFontString()
			displayBar.spLabel:SetFont("Fonts\\FRIZQT__.TTF", 18, "OUTLINE")
			displayBar.spLabel:SetPoint("TOPLEFT",50,0)
			displayBar.spLabel:SetPoint("BOTTOMRIGHT",displayBar,"BOTTOMLEFT",150,0)
			displayBar.hasteLabel = displayBar:CreateFontString()
			displayBar.hasteLabel:SetFont("Fonts\\FRIZQT__.TTF", 18, "OUTLINE")
			displayBar.hasteLabel:SetPoint("TOPLEFT",150,0)
			displayBar.hasteLabel:SetPoint("BOTTOMRIGHT")
			display.bars[num] = displayBar
		else
			displayBar:Show()
		end
		local key, tex = next(dotTable,lastKey)
		displayBar.icon:SetTexture(tex)
		displayBar.spellID = key
		lastKey = key
	end
	for num=keyCount+1,#display.bars do
		display.bars[num]:Hide()
	end
end
display:SetScript("OnEvent", function(self,event,...)
	self.pguid = self.pguid or UnitGUID("player")
	if event == "SPELLS_CHANGED" then
		self:UnregisterEvent("SPELLS_CHANGED")
		init()
	elseif (event == "ADDON_LOADED") and ((...) == "DotHaste") then
		DotHaste_Saved = DotHaste_Saved or {p="CENTER",r="CENTER",n="UIParent",x=0,y=0,s=1,lock=false,}
		self.saved = DotHaste_Saved
		display:SetScale(self.saved.s*0.5)
		display:UpdateAnchor()
		display:UpdateLock()
		self.saved.breakpoints = self.saved.breakpoints or {}
		for key, breakpoints in pairs(dotBreakpoints[(select(2,UnitClass("player")))] or {}) do
			self.saved.breakpoints[key] = self.saved.breakpoints[key] or breakpoints
		end
		self:UnregisterEvent("ADDON_LOADED")
	elseif (event == "COMBAT_LOG_EVENT_UNFILTERED") then
		if select(4,...) == self.pguid then
			local subEvent = select(2,...)
			if (subEvent == "SPELL_AURA_APPLIED") or (subEvent == "SPELL_AURA_REFRESH") then
				local spellID = select(12,...)
				spellID = mappedSpellIDs[spellID] or spellID
				if dotTable[spellID] then
					local sp, hastePct = spelldamage(), UnitSpellHaste("player")
					local nextBreakpoint = math.huge
					for _,value in ipairs(self.saved.breakpoints[spellID] or {}) do
						if value > hastePct then
							nextBreakpoint = value
							break
						end
					end
					local targetGUID = select(8,...)
					dotDataSp[spellID][targetGUID] = sp
					dotDataHaste[spellID][targetGUID] = hastePct
					dotDataBreakpoint[spellID][targetGUID] = nextBreakpoint
					display:Update()
				end
			elseif (subEvent == "SPELL_AURA_REMOVED") then
				local spellID = select(12,...)
				spellID = mappedSpellIDs[spellID] or spellID
				if dotTable[spellID] then
					local dstGUID = select(8,...)
					dotDataSp[spellID][dstGUID] = nil
					dotDataHaste[spellID][dstGUID] = nil
					dotDataBreakpoint[spellID][dstGUID] = nil
					display:Update()
				end
			end
		end
	elseif (event == "UNIT_SPELL_HASTE") and UnitIsUnit((...),"player") then
		display:Update()	
	elseif (event == "PLAYER_TALENT_UPDATE") or (event == "ACTIVE_TALENT_GROUP_CHANGED") then
		talentupdate()
	else
		display:Update()
	end
end)
display:RegisterEvent("ADDON_LOADED")
display:RegisterEvent("PLAYER_TALENT_UPDATE")
display:RegisterEvent("ACTIVE_TALENT_GROUP_CHANGED")
LibStub("LibShorterSlash-1.0"):RegisterCommand(slashTable,"/dothaste","/dh")